/*global global */ // <-- jshint
/*
 * Wishlist:
 * - provide more information for functions in GI repositories -
 *   e.g. Meta.Display.begin_grab_op
 *   This is almost done in that if you can get the BaseInfo corresponding to
 *   the function in question, then you can use 'describeCallable' (already
 *   coded in this file) to produce a string with not only the function
 *   code, but information about the arguments.
 *
 *   The problem is that given a function we cannot work out its
 *   namespace/class: for example
 *     imports.gi.GIRepository.Repository.get_default.toString() ==
 *     'function get_default(){
 *          / * proxy for native symbol g_irepository_get_default(); * /
 *      }'
 *
 *   Using regex, we can work out that a function belongs to a GI repository
 *   by looking for 'proxy for native symbol', and we can get the function name
 *   via 'function [function_name]', BUT the problem is determining both
 *   'GIRepository' and 'Repository', since the way we look up a method is by:
 *
 *       rep = GIR.Repository.get_default();
 *       info = rep.find_by_name('GIRepository', 'Repository'); // <- need this!
 *       funcInfo = GIR.object_info_find_method(info, 'get_default');
 *
 *   So, how do we determine the 'GIRepository' and 'Repository' from just the
 *   function?
 *
 *   UPDATE: don't think this is possible. Some harder examples:
 *   g_irepository_get_default -> GIRepository.Repository get_default
 *   gdk_x11_display_error_trap_pop -> GdkX11.X11Display error_trap_pop
 *   gdk_x11_atom_to_xatom -> GdkX11.x11_atom_to_xatom
 *   FT_Library_Version -> freetype2.library_version // <-- !!!
 *
 *   There's a function rep.get_c_prefix(namespace) that returns the C prefix
 *   for that namespace, but the function names are *NOT* consistent with this
 *   (for example GIRepository's C prefix is 'GI' but all the function names
 *   are prefixed by 'g_').
 *
 * TESTS:
 * Meta
 * Meta.Display --> should show function code + 'inspect prototype'
 * global.screen --> should show 'inspect prototype' + 'run_dispose' and
 *                   'notify'
 * imports --> "[object GjsFileImporter']" (not on GNOME 3.2: this will
 *  fail with 'No JS module 'actor' found in search path: even with the default
 *  looking glass).
 * imports.gi --> browseable
 *
 * Clutter.Event (InfoType.UNION)
 * Meta.Rectangle (InfoType.STRUCT)
 * Gtk.Orientable (InfoType.INTERFACE) (not on GNOME 3.2: "API of type
 *  INTERFACE not implemented, cannot define ABC.XYZ")
 */

/// Limit the maximum levels of 'back' you get (in default looking glass it's 1)
/// It's just here to stop the history getting unreasonably large and probably
/// (hopefully?) will never be actually attained.
const MAX_BACK_LEVELS = 100;

/// impose a limit on the number of lines in the looking glass when inspecting
///  objects before a 'more...' button becomes available.
/// (if you inspect Gtk --> Classes you'll see what I mean).
const ITEM_LIMIT = 80;

/* **** CODE ***** */
// imports
const Clutter = imports.gi.Clutter;
const GIR = imports.gi.GIRepository;
const Lang = imports.lang;
const Mainloop = imports.mainloop;
const Pango = imports.gi.Pango;
const Shell = imports.gi.Shell;
const St = imports.gi.St;

const ExtensionSystem = imports.ui.extensionSystem;
const LookingGlass = imports.ui.lookingGlass;
const Main = imports.ui.main;
const PopupMenu = imports.ui.popupMenu;

const Gettext = imports.gettext.domain('looking-glass-helper');
const _ = Gettext.gettext;
const GJS_ = imports.gettext.domain('gjs').gettext;

/// skip the following types
/// (could include GIR.InfoType.INVALID, INVALID_0, UNRESOLVED, ...)
const SKIP = {};
SKIP[GIR.InfoType.CALLBACK] = true;

// laziness
const objectToString = LookingGlass.objectToString;
const ObjLink = LookingGlass.ObjLink;
const OIProto = LookingGlass.ObjInspector.prototype;
const EProto = LookingGlass.Extensions.prototype;
const IProto = LookingGlass.Inspector.prototype;
const OLProto = ObjLink.prototype;

// compatibility: GIR.base_info_get_type(info) (3.2) vs info.get_type (3.4)
if (!GIR.base_info_get_type) {
    GIR.base_info_get_type = function (info) {
        return info.get_type();
    };
}

// pre-caching regex
const NAMESPACE_REGEX = '([A-Za-z][A-Za-z0-9_]*)';
const CLASS_REGEX = '([A-Za-z][A-Za-z0-9_]*)';

// Note: for the GI-related regexes ('[GIName:<repo> ..',
// '/* proxy for native symbol ... */',) gjs sources: gi/object.c,
// gi/function.c.

// imports.gi.(namespace)
const IMPORTS_GI_REGEX = new RegExp('imports\\.gi\\.' + NAMESPACE_REGEX);
// ^(namespace)$
const NAMESPACE_ONLY_REGEX = new RegExp('^' + NAMESPACE_REGEX + '$');
// GIName:Namespace.Classname jsobj
const INSTANCE_OR_CLASS_REGEX = new RegExp('GIName:' + NAMESPACE_REGEX + '\\.' +
        CLASS_REGEX + ' jsobj');
// [object _private_Meta_Rectangle] for Structs or Unions
const INSTANCE_OR_CLASS_REGEX_PRIVATE = new RegExp('^\\[object _private_' +
        NAMESPACE_REGEX + '_' + CLASS_REGEX + '\\]$');
// function _private_Namespace_Method() {\s+[native code]
const CONSTRUCTOR_REGEX = new RegExp('^function _private_' + NAMESPACE_REGEX +
        '_' + CLASS_REGEX + '\\(\\) \\{\\s+\\[native code\\]');

// e.g. meta_screen_append_new_workspace -> Meta.Screen append_new_workspace
// problems (examples):
// g_irepository_get_default -> GIRepository.Repository get_default
// gdk_x11_display_error_trap_pop -> GdkX11.X11Display error_trap_pop
// gdk_x11_atom_to_xatom -> GdkX11.x11_atom_to_xatom
// NOTE: in future versions of GJS they put the arguments in here.
// "function %s(%s) {\n\t/* proxy for native symbol %s(); */\n}"
// "function %s(%s) {\n\t/* proxy for native symbol */\n}" for callables that
//                                                         are not functions.
const GI_FUNCTION_REGEX = new RegExp('^function ([a-zA-Z][a-zA-Z0-9_]*)\\(\\)' +
    '\\{\\s+/\\* proxy for native symbol ' +
    '([a-zA-Z][a-zA-Z0-9]*(?:_[a-zA-Z0-9]+)*)' + // prefix: gdx_x11, meta, ...
    '_\\1\\(');

const TRANSLATIONS = {};

TRANSLATIONS[GIR.InfoType.INVALID] = _("Other");
TRANSLATIONS[GIR.InfoType.FUNCTION] = _("Functions"); // 1
TRANSLATIONS[GIR.InfoType.CALLBACK] = _("Functions"); // 2
TRANSLATIONS[GIR.InfoType.STRUCT] = _("Classes"); // 3 (? e.g. Meta.Strut)
TRANSLATIONS[GIR.InfoType.BOXED] = _("Other"); // 4  ???????????????????
TRANSLATIONS[GIR.InfoType.ENUM] = _("Constants"); // 5
TRANSLATIONS[GIR.InfoType.FLAGS] = _("Constants"); // 6
TRANSLATIONS[GIR.InfoType.OBJECT] = _("Classes"); // 7 (e.g. Meta.Screen)
TRANSLATIONS[GIR.InfoType.INTERFACE] = _("Interfaces"); // 8 Gtk.Orientable
TRANSLATIONS[GIR.InfoType.CONSTANT] = _("Constants"); // 9
TRANSLATIONS[GIR.InfoType.INVALID_0] = _("Other"); // 10
TRANSLATIONS[GIR.InfoType.UNION] = _("Classes"); // 11 (e.g. Clutter.Event)
TRANSLATIONS[GIR.InfoType.VALUE] = _("Constants"); // 12 enum value of an enum
TRANSLATIONS[GIR.InfoType.SIGNAL] = _("Signals"); // 13
TRANSLATIONS[GIR.InfoType.VFUNC] = _("Virtual Functions"); // 14
TRANSLATIONS[GIR.InfoType.PROPERTY] = _("Properties"); // 15
TRANSLATIONS[GIR.InfoType.FIELD] = _("Properties"); // 16
TRANSLATIONS[GIR.InfoType.ARG] = _("Arguments"); // 17
TRANSLATIONS[GIR.InfoType.TYPE] = _("Other"); // 18 ??????
TRANSLATIONS[GIR.InfoType.UNRESOLVED] = _("Other"); // 19

/// These are fields that we can query .get_n_[Field] of.
const Fields = {
    ARG: 0,
    CONSTANT: 1,
    FIELD: 2,
    INFO: 3,
    INTERFACE: 4,
    METHOD: 5,
    PREREQUISITE: 6,
    PROPERTY: 7,
    SIGNAL: 8,
    VALUE: 9,
    VFUNC: 10
};

/// The first element is the 'get_%s(info, i)' method, the second is
/// 'get_n_%s(info)' (e.g. 'property' for get_%s and 'properties' for get_n_%s)
const _FieldSingularPlurals = {};
_FieldSingularPlurals[Fields.ARG] = ['arg', 'args'];
_FieldSingularPlurals[Fields.CONSTANT] = ['constant', 'constants'];
_FieldSingularPlurals[Fields.FIELD] = ['field', 'fields'];
_FieldSingularPlurals[Fields.INFO] = ['info', 'infos'];
_FieldSingularPlurals[Fields.INTERFACE] = ['interface', 'interfaces'];
_FieldSingularPlurals[Fields.METHOD] = ['method', 'methods'];
_FieldSingularPlurals[Fields.PREREQUISITE] = ['prerequisite', 'prerequisites'];
_FieldSingularPlurals[Fields.PROPERTY] = ['property', 'properties'];
_FieldSingularPlurals[Fields.SIGNAL] = ['signal', 'signals'];
_FieldSingularPlurals[Fields.VALUE] = ['value', 'values'];
_FieldSingularPlurals[Fields.VFUNC] = ['vfunc', 'vfuncs'];

/// This defines the fields that we are interested in for each XYZ in
///  XYZ_info_get_n_[field]. Use get_prefix_for_n(info) to get the XYZ for
/// that info type (for example GIR.InfoType.SIGNAL and GIR.InfoType.FUNCTION
/// both return prefix 'callable' and the only applicable method is
/// 'callable_info_get_n_arg'.
const FieldsOfInterest = {
    callable: [Fields.ARG],
    struct: [Fields.FIELD, Fields.METHOD],
    union: [Fields.FIELD, Fields.METHOD],
    //'enum': [], // skip (GJS converts to an enum)
    object: [Fields.CONSTANT, Fields.FIELD, Fields.INTERFACE, Fields.METHOD,
             Fields.PROPERTY, Fields.SIGNAL, Fields.VFUNC],
    'interface': [Fields.CONSTANT, Fields.METHOD, Fields.PROPERTY,
                  Fields.SIGNAL, Fields.VFUNC]
};

/* Global variables */
let storage = {}, rep = null, namespaceProxies = {}, objectProxies = {};
let uuid;
const GIRepository = {};

/*******
 * Helper functions
 *******/
function debug(message) {
    log(message);
}

function warn(message) {
    log(_("[LG-helper] WARNING: ") + message);
}

function error(message) {
    // todo: there's some sort of throwExtensionError in extensionSystem right?
    log(_("[LG-helper] ERROR: !!! ") + message);
    ExtensionSystem.logExtensionError(uuid, message);
}


/**
 * Performs obj[property] in a way that is robust to `obj` being (eg) `imports`,
 * in which you'd get an error 'No JS module 'property' found on the search
 * path'.
 *
 * If `obj` has property `property`, it returns the array
 *     [true, obj[property]]
 *
 * Otherwise it will return null.
 *
 * Hence you can use this as a check to see if the property is even defined:
 *
 * let val = get_property(object, propName);
 * if (val) {
 *     val = val[1];
 * } else {
 *     print("property '%s' does not exist".format(propName));
 * }
 *
 * This is to distinguish from cases where the property *is* defined but
 * has its value set to (say) `undefined`.
 */
function get_property(obj, property) {
    let keys = Object.getOwnPropertyNames(obj); // includes non-enumerable
    if (keys.indexOf(property) > -1) {
        return [true, obj[property]];
    }
    return null;
}

/*
 * This function tries to guess whether a given object (function?) is a
 * constructor for a class with a .prototype() that we might want to explore
 * in the looking glass. It's all a bit confusing to me, but basically if
 * I inspect a function that 'is interesting', it should provide me a link
 * to explore that class' prototype.
 *
 * Some test cases:
 * Meta.Display should be interesting
 * - prototype.constructor: 'function _private_Meta_Display() { [native code] }'
 * - prototype: 'object prototype of GIName:..'
 * PopupMenu.PopupMenu should be interesting.
 * - prototype.constructor: 'function Object() { [native code] }'
 * - prototype: 'object Object'
 *
 * function () {} should *not* be interesting
 * - prototype.constructor: 'function () {}'
 * - prototype: 'object Object'
 * PopupMenu.PopupMenu.prototype.addMenuItem should *not* be interesting.
 * - prototype.constructor: 'function (menuItem, position) {...'
 * - prototype: 'object Object'
 * PopupMenu.PopupMenu.prototype.connect should *not* be interesting.
 * - prototype.constructor: 'function _connect(name, callback) {...'
 * - prototype: 'object Object'
 *
 * So what distinguishes these ????
 * Could do a hack: look for '.prototype._init' for gnome-shell classes and
 * '_private_[namespace]_[classname]' for GIR classes ?
 */
function isInterestingClass(obj) {
    let isClass = (obj && typeof(obj) === 'function' && obj.prototype &&
            (obj.prototype._init || // gnome-shell classes
             CONSTRUCTOR_REGEX.test(obj.toString())));
    if (isClass) {
        try {
            // some classes like Meta.BackgroundActor.prototype have
            // cause problems when you try to .toString() them, so we won't
            // provide any ObjLinks to those.
            obj.prototype.toString();
            return true;
        } catch (e) {
        }
    }
    return false;
}

/* tests if a imports.gi[ns] exists and is equal to `obj` */
function testNamespaceAgainst(ns, obj) {
    let result = false;
    try {
        result = (imports.gi[ns] === obj);
    } catch (error) {
        result = false;
    }
    return result;
}

/*
 * This function takes in an object `obj` that is a GIRepository Namespace and
 * attempts to match is to a namespace (i.e. imports.gi[namespace] === obj).
 *
 * This is only used when a namespace that was not loaded when the extension
 * was loaded becomes loaded (e.g. imports.gi.GdkX11).
 *
 * It is tied to the looking glass, i.e. expected to be called after the user
 * has typed 'imports.gi.GdkX11' into the LG and we are trying to match up
 * the resulting object to get 'GdkX11' back out.
 *
 * It does the guessing by:
 *
 * 1. look at the most recent looking glass command for 'imports.gi.GdkX11'
 * 2. look at the most recent looking glass command for 'GdkX11'
 * 3. failing that, loop through all loaded namespaces and match against `obj`.
 *
 *
 * It returns 'null' if it couldn't guess it.
 */
function guessNamespace(obj) {
    let hist = Main.lookingGlass._history._history,
        cmd = hist[hist.length - 1],
        match = null,
        ns = null;
    if (cmd) {
        cmd = cmd.trim();
        // 1. try imports.gi.[namespace]
        if ((match = IMPORTS_GI_REGEX.exec(cmd)) &&
                testNamespaceAgainst(match[1], obj)) {
            ns = match[1];
        } else {
            // 2. say they just typed in 'Meta'
            // (unlikely, have to imports.gi.XXX first)
            if (NAMESPACE_ONLY_REGEX.exec(cmd) &&
                    testNamespaceAgainst(cmd, obj)) {
                ns = cmd;
            }
        }
    }
    // 3. last-ditch effort: it's loaded now, so loop through and compare
    if (!ns) {
        let namespaces = rep.get_loaded_namespaces();
        for (let i = 0; i < namespaces.length; ++i) {
            if (imports.gi[namespaces[i]] === obj) {
                ns = namespaces[i];
                break;
            }
        }
    }
    if (ns) {
        GIRepository[ns] = imports.gi[ns];

        Object.defineProperty(GIRepository[ns], '__LG_helper_name__', {
            value: ns,
            enumerable: false,
            writable: false
        });
    }

    return ns;
}

/**
 * Get's an info's prefix according to g_PREFIX_info_get_n_args etc
 * If there is no relevant get_n_yyy function then it returns null.
 */
function get_prefix_for_n(info) {
    let type = GIR.base_info_get_type(info);
    switch (type) {
    /* CallableInfo */
    // FunctionInfo
    case GIR.InfoType.FUNCTION:
    case GIR.InfoType.CALLBACK:

    // SignalInfo
    case GIR.InfoType.SIGNAL:

    // VFuncInfo
    case GIR.InfoType.VFUNC:
        return 'callable';

    // StructInfo
    case GIR.InfoType.BOXED:
    case GIR.InfoType.STRUCT:
        return 'struct';

    // UnionInfo
    case GIR.InfoType.UNION:
        return 'union';

    // EnumInfo
    case GIR.InfoType.ENUM:
        return 'enum';

    // ObjectInfo
    case GIR.InfoType.OBJECT:
        return 'object';

    // InterfaceInfo
    case GIR.InfoType.INTERFACE:
        return 'interface';

    // ConstantInfo
    case GIR.InfoType.CONSTANT:
    // ValueInfo
    case GIR.InfoType.FLAGS:
    case GIR.InfoType.VALUE:
    // PropertyInfo
    case GIR.InfoType.PROPERTY:
    // FieldInfo
    case GIR.InfoType.FIELD:
    // TypeInfo
    case GIR.InfoType.TYPE:
    // Other
    case GIR.InfoType.INVALID:
    case GIR.InfoType.INVALID_0:
        /* falls through */
    default:
        return null;
    }

    return null;
}

/* Helper function - just a wrapper around the `GIR.XYZ_info_get_n_ABC`
 * functions (It's a pain to re-write heaps of code just changing the XYZ
 * depending on the info type)
 */
function get_n(info, field) {
    let plural = _FieldSingularPlurals[field];
    if (plural) {
        let prefix = get_prefix_for_n(info);
        if (prefix) {
            let func_name = '%s_info_get_n_%s'.format(prefix, plural[1]);
            if (GIR[func_name]) {
                return GIR[func_name](info);
            }
        }
    }
    return 0;
}

/* Helper function - just a wrapper around the `GIR.XYZ_info_get_ABC` functions.
 * (It's a pain to re-write heaps of code just changing the XYZ depending on the
 * info type)
 */
function get_field(info, field, i) {
    let prefix = get_prefix_for_n(info),
        singular = _FieldSingularPlurals[field][0];

    if (prefix) {
        let func_name = '%s_info_get_%s'.format(prefix, singular);
        if (GIR[func_name]) {
            return GIR[func_name](info, i);
        }
    }
    return null;
}


/* Convenience function converting GIR.TypeInfo to a string */
function typeInfoToString(typeInfo) {
    if (typeInfo) {
        let tag = GIR.type_info_get_tag(typeInfo);
        if (tag === GIR.TypeTag.INTERFACE) {
            // get the further type
            let info = GIR.type_info_get_interface(typeInfo);
            return info.get_name();
        } else {
            return GIR.type_tag_to_string(tag);
        }
    }
    return _("unknown");
}

/* convenience function: retrieves the property from an object if it exists
 * and otherwise gets its type
 */
function getValueOrDescription(object, prop, type, info, namespaceObj) {
    let val = prop;
    try {
        val = object[prop];
        if (val === undefined) {
            // e.g. if object is a prototype
            switch (type) {
            case GIR.InfoType.INTERFACE:
                val = namespaceObj[prop];
                break;
            case GIR.InfoType.CONSTANT:
                val = _("type: ") +
                    typeInfoToString(GIR.constant_info_get_type(info));
                break;
            case GIR.InfoType.FIELD:
                val = _("type: ") +
                    typeInfoToString(GIR.field_info_get_type(info));
                break;
            case GIR.InfoType.PROPERTY:
                val = _("type: ") +
                    typeInfoToString(GIR.property_info_get_type(info));
                break;
            case GIR.InfoType.ARG:
                val = _("type: ") +
                    typeInfoToString(GIR.arg_info_get_type(info));
                val += ' (' + (GIR.arg_info_is_return_value(info) ?
                    _("output") : _("input"));
                val += GIR.arg_info_is_optional(info) ?
                    ', ' + _("optional") : '';
                val += GIR.arg_info_may_be_null(info) ?
                    ', ' + _("may be null") : '';
                val += GIR.arg_info_may_be_null(info) ?
                    ', ' + _("caller allocates") : '';
                val += ')';
                break;
            case GIR.InfoType.VALUE:
                val = GIR.value_info_get_value(info);
                break;
            /*
            case GIR.InfoType.SIGNAL:
                break;
            */
            default:
                val = prop;
                break;
            }
        }
    } catch (error) {
        let pInfo = info.get_container(),
            pName = (pInfo ? pInfo.get_name() : '');
        val = prop;

        if (error.message === GJS_("Can't get field %s.%s from a prototype").format(
                    pName, prop)) {
            // e.g. Meta.Rectangle.prototype.x: "Can't get field Rectangle.x
            // from a prototype" --> retrieve the type.
            val = getValueOrDescription({}, prop, type, info, namespaceObj);
        } else if (type !== GIR.InfoType.INTERFACE) {
            // On GNOME 3.2 INTERFACE is not implemented (on 3.4 it's fine);
            // skip those warnings.
            if (pName) {
                // e.g. Meta.Display.begin_grab_op
                warn(_("Error loading namespace member %s.%s.%s: %s").format(
                            info.get_namespace(), pName, prop, error.message));
            } else {
                // e.g. Meta.Display
                warn(_("Error loading namespace member %s.%s: %s").format(
                            info.get_namespace(), prop, error.message));
            }
        }
    }
    return val;
}

/* Describes a callback/signal/function (GICallableInfo) in a string. */
// Currently not used because we have no createFunctionProxy(func)
// (though the function itself works nicely!)
function describeCallable(baseInfo) {
    let name = baseInfo.get_name(),
        desc = name + '(',
        nargs = GIR.callable_info_get_n_args(baseInfo);
    for (let i = 0; i < nargs; ++i) {
        let info = GIR.callable_info_get_arg(baseInfo, i);
        desc += info.get_name();
        desc += ' /* ' + getValueOrDescription({}, '', GIR.InfoType.ARG, info) +
                    ' */';
        if (i !== (nargs - 1)) {
            desc += ',\n    ';
        }
    }
    desc += '\n)';
    desc += '\nReturns: ' +
        typeInfoToString(GIR.callable_info_get_return_type(baseInfo));
    if (GIR.callable_info_may_return_null(baseInfo)) {
        desc += ' (may return null)';
    }
    return desc;
}

/* loads properties by type */
function _loadByType(info_, field, obj) {
    let n = get_n(info_, field),
        container = {},
        labels = '',
        namespaceObj = imports.gi[info_.get_namespace()];

    for (let i = 0; i < n; ++i) {
        let info = get_field(info_, field, i),
            name = info.get_name(),
            type = GIR.base_info_get_type(info);
        labels += name;
        if (i !== (n - 1)) {
            labels += ', ';
        }
        container[name] = getValueOrDescription(obj, name, type, info,
                namespaceObj);
    }
    if (n <= 0) {
        return null;
    }
    Object.defineProperty(container, '__LG_helper_label__', {
        value: labels,
        enumerable: false,
        writable: true
    });
    return container;
}

/*
 * When called with a namespace (e.g. "Meta") and its object (imports.gi.Meta),
 * will return an object representing a browseable menu of that namespace.
 *
 * It will have members 'Interfaces', 'Functions', 'Constants' (etc) that are
 * all objects with the given interfaces/functions/constants.
 *
 * The main reason of having the nested object structure is that when this gets
 * inspected in the Looking Glass, it can be browsed just like any normal
 * object.
 */
function createNamespaceProxy(namespace, namespaceObj) {
    if (namespaceProxies[namespace]) {
        return namespaceProxies[namespace];
    }
    debug('createNamespaceProxy(%s, %s)'.format(
        namespace, namespaceObj.toString())
    );
    let n_info = rep.get_n_infos(namespace),
        proxy = {};
    namespaceObj = namespaceObj || imports.gi[namespace];

    if (n_info < 0) {
        warn(_("Could not locate namespace %s").format(namespace));
    }

    // load members
    for (let i = 0; i < n_info; ++i) {
        let info = rep.get_info(namespace, i),
            type = GIR.base_info_get_type(info),
            name = info.get_name(),
            class_ = TRANSLATIONS[type] || _("Other"),
            childObj = proxy[class_];

        if (SKIP[type]) {
            continue;
            // for now, callbacks (e.g. Meta.PrefsChangedFunc)
        }
        if (childObj === undefined) {
            proxy[class_] = {};
            childObj = proxy[class_];
            Object.defineProperty(childObj, '__LG_helper_label__', {
                value: '',
                enumerable: false,
                writable: true
            });
        }
        childObj[name] = getValueOrDescription(namespaceObj, name, type,
                info, namespaceObj);
        childObj.__LG_helper_label__ += name + ', ';
    }
    for (let class_ in proxy) {
        if (proxy.hasOwnProperty(class_)) {
            // remove trailing comma
            proxy[class_].__LG_helper_label__.replace(/, *$/, '');
        }
    }

    namespaceProxies[namespace] = proxy;
    return proxy;
}


/*
 * When called with a namespace (e.g. "Meta"), a class (e.g. "Display") and an
 * object (Meta.Display.prototype, global.display), this function will return
 * an object representing a browseable menu of that class's methods, signals
 * and properties.
 *
 * It will have members 'Functions', 'Signals' (etc) that are all arrays with
 * the given interfaces/functions/constants.
 *
 * The main reason of having the nested structure is that when this gets
 * inspected in the Looking Glass, it can be browsed just like any normal
 * object.
 *
 * `baseObject` can be an instance or a prototype.
 *
 * This is *meant* to be a bit more generalised in that the type of the
 * namespace.classname being inspected can be any of:
 * - GIR.InfoType.OBJECT (most classes of a namespace)
 * - GIR.InfoType.STRUCT (e.g. Meta.Rectangle)
 * - GIR.InfoType.BOXED (e.g. ??)
 * - GIR.InfoType.UNION (e.g. Clutter.Event)
 * - GIR.InfoType.INTERFACE (e.g. Gtk.0rientable) <-- TODO
 */
function createObjectProxy(namespace, classname, baseObject) {
    let id = namespace + '.' + classname,
        proxy = {},
        info = rep.find_by_name(namespace, classname);
    if (!info) {
        warn(_("Could not locate object %s").format(id));
        return null;
    }
    let prefix = get_prefix_for_n(info);
    if (!FieldsOfInterest[prefix]) {
        warn(_("Namespace member %s is of type '%sInfo' and should not have " +
                    "createObjectProxy called on it").format(id, prefix));
        return null;
    }
    debug('createObjectProxy(%s, %s) (%sInfo)'.format(
                namespace, classname, prefix));

    if (objectProxies[id]) {
        proxy = objectProxies[id];
    } else {
        // retrieve the parent information if it's there.
        if (prefix === 'object') {
            let parentInfo = GIR.object_info_get_parent(info),
                label = '';
            if (parentInfo) {
                let parentNS = parentInfo.get_namespace(),
                    parentName = parentInfo.get_name(),
                    parentClass = null;
                if (parentNS !== '') {
                    label = parentNS + '.' + parentName;
                } else {
                    label = parentName;
                }

                try {
                    parentClass = imports.gi[parentNS][parentName];
                } catch (error) {
                    // non-loaded or non-existent parent namespace..
                }
                if (!parentClass) {
                    parentClass = label;
                }
                proxy[_("Parent") + ' (' + label + ')'] = parentClass;
            }
        }
        proxy[_("Interfaces")] = _loadByType(info, Fields.INTERFACE, baseObject);
        proxy[_("Preqrequisites")] = _loadByType(info, Fields.PREREQUISITE, baseObject);
        proxy[_("Constants")] = _loadByType(info, Fields.CONSTANT, baseObject);
        proxy[_("Methods")] = _loadByType(info, Fields.METHOD, baseObject);
        proxy[_("Signals")] = _loadByType(info, Fields.SIGNAL, baseObject);
        proxy[_("Virtual Functions")] = _loadByType(info, Fields.VFUNC, baseObject);
        proxy[_("Arguments")] = _loadByType(info, Fields.ARG, baseObject);
        proxy[_("Enum Values")] = _loadByType(info, Fields.VALUE, baseObject);
    }

    // if `baseObject` is an instance, properties and fields can change
    // from instance-to-instance.
    proxy[_("Properties")] = _loadByType(info, Fields.PROPERTY, baseObject);

    let tmp = _loadByType(info, Fields.FIELD, baseObject);
    if (tmp) {
        if (proxy[_("Properties")]) {
            proxy[_("Properties")].__LG_helper_label__ += ', ';
        }
        for (let f in tmp) {
            if (tmp.hasOwnProperty(f)) {
                if (!proxy[_("Properties")]) {
                    proxy[_("Properties")] = {};
                    Object.defineProperty(proxy[_("Properties")],
                        '__LG_helper_label__', {
                            value: '',
                            enumerable: false,
                            writable: true
                        }
                    );
                }
                proxy[_("Properties")][f] = tmp[f];
            }
        }
        proxy[_("Properties")].__LG_helper_label__ += tmp.__LG_helper_label__;
    }

    // delete null fields
    for (let cat in proxy) {
        if (proxy.hasOwnProperty(cat) && !proxy[cat]) {
            delete proxy[cat];
        }
    }

    if (!objectProxies[id]) {
        // if this is the first-time creation of the object, we add a list
        // of all the in-built property names from the GIR file to the proxy
        Object.defineProperty(proxy, '__LG_helper_inbuilt_props__', {
            value: [],
            enumerable: false,
            writable: true
        });

        for (let type in proxy) {
            if (!proxy.hasOwnProperty(type)) {
                continue;
            }
            for (let prop in proxy[type]) {
                if (!proxy[type].hasOwnProperty(prop)) {
                    continue;
                }
                proxy.__LG_helper_inbuilt_props__.push(prop);
            }
        }
    }

    // if `baseObject` is an instance it may have other things added in...
    // (e.g. global.screen.workspace_grid): add these in to the proxy.
    // Note - I don't use .hasOwnProperty because selectObject doesn't so I
    // want the same results as if I had used the .selectObject function.
    let builtin = proxy.__LG_helper_inbuilt_props__,
        any = false;
    for (let prop in baseObject) {
        if (builtin.indexOf(prop) === -1) {
            if (!any) {
                any = true;
                proxy.__LG_helper_other_properties__ = {
                    __LG_helper_nolink__: true,
                    widget: new St.Label({
                        text: '\n' + _("Other properties") + ':'
                    })
                };
            }
            proxy[prop] = baseObject[prop];
        }
    }

    objectProxies[id] = proxy;
    return proxy;

}

function Inspector() {
    this._init();
}

Inspector.prototype = {
    __proto__: IProto,

    _init: function () {
        // there is no this._eventHandler.disconnectAll() so instead of
        // calling the parent method we'll duplicate it here without
        // the 'connect' code.
        let container = new Shell.GenericContainer({ width: 0,
                                                     height: 0 });
        container.connect('allocate', Lang.bind(this, this._allocate));
        Main.uiGroup.add_actor(container);

        let eventHandler = new St.BoxLayout({ name: 'LookingGlassDialog',
                                              vertical: false,
                                              reactive: true });
        this._eventHandler = eventHandler;
        Main.pushModal(this._eventHandler);
        container.add_actor(eventHandler);
        this._displayText = new St.Label();
        eventHandler.add(this._displayText, { expand: true });

        this._borderPaintTarget = null;
        this._borderPaintId = null;
        eventHandler.connect('destroy', Lang.bind(this, this._onDestroy));
        // skip all the eventHandler.connect('*-event', ...); in favour of:
        this._capturedEventId = global.stage.connect('captured-event',
                Lang.bind(this, this._onCapturedEvent));
        // skip the pointer/keyboard grab

        // this._target is the actor currently shown by the inspector.
        // this._pointerTarget is the actor directly under the pointer.
        // Normally these are the same, but if you use the scroll wheel
        // to drill down, they'll diverge until you either scroll back
        // out, or move the pointer outside of _pointerTarget.
        this._target = null;
        this._pointerTarget = null;
    },

    /* dispatch events */
    _onCapturedEvent: function (actor, event) {
        let type = event.type(),
            mod = event.get_state();
        // if Shift is on, pass through the event.
        if (mod & Clutter.ModifierType.SHIFT_MASK) {
            return false;
        }
        switch (type) {
        case Clutter.EventType.KEY_PRESS:
            return this._onKeyPressEvent(actor, event);
        case Clutter.EventType.BUTTON_PRESS:
            return this._onButtonPressEvent(actor, event);
        case Clutter.EventType.SCROLL:
            return this._onScrollEvent(actor, event);
        case Clutter.EventType.MOTION:
            return this._onMotionEvent(actor, event);
        default:
            return false;
        }
    },

    _close: function () {
        global.stage.disconnect(this._capturedEventId);
        Main.popModal(this._eventHandler);
        IProto._close.call(this);
    }

};

function ObjLinkWithRightClick(o, title) {
    this._init(o, title);
}
ObjLinkWithRightClick.prototype = {
    __proto__: OLProto,

    _init: function (o, title) {
        OLProto._init.call(this, o, title);

        this.actor.set_button_mask(St.ButtonMask.ONE | St.ButtonMask.THREE);

        // this is what gets copied when the user clicks 'copy' (as opposed to
        // objectToString(o)).
        this._text = '' + o;

        // clipboard
        this._clipboard = St.Clipboard.get_default();
    },

    // we only create the menu on demand.
    _createRightClickMenu: function () {
        // make a right click menu (copy off _EntryMenu but without 'paste' or
        // 'password' options)
        this.menu = new PopupMenu.PopupMenu(this.actor, 0, St.Side.TOP);
        this.menu.actor.add_style_class_name('entry-context-menu');

        let item;
        item = new PopupMenu.PopupMenuItem(_("Copy as string"));
        item.connect('activate', Lang.bind(this, this._onCopyActivated));
        this.menu.addMenuItem(item);

        item = new PopupMenu.PopupMenuItem(_("Insert"));
        item.connect('activate', Lang.bind(this, this._onInsertActivated));
        this.menu.addMenuItem(item);

        Main.uiGroup.add_actor(this.menu.actor);
        this.menu.actor.hide();

        this._menuManager = new PopupMenu.PopupMenuManager({actor: this.actor});
        this._menuManager.addMenu(this.menu);
    },

    _onCopyActivated: function () {
        this._clipboard.set_text(this._text);
    },

    _onInsertActivated: function () {
        Main.lookingGlass.insertObject(this._obj);
    },

    // launch the menu or the object inspector (button is a Clutter button
    // number: 1 is primary, 3 is right-click)
    _onClicked: function (link, button) {
        if (this.menu && this.menu.isOpen) {
            this.menu.close();
        } else if (button === 1) { // left click
            OLProto._onClicked.call(this);
        } else { // right click
            if (!this.menu) {
                this._createRightClickMenu();
            }
            this.menu.open();
        }
    }
};

/* *********** EXTENSION **************** */
function init(metadata) {
    namespaceProxies = {}; // these persist if you enable/disable multiple times
                           // in one gnome-shell session.
    objectProxies = {};
    uuid = metadata.uuid;
}

function enable() {
    // TODO: enable after a disable? Looking Glass already exists.

    rep = GIR.Repository.get_default();
    let namespaces = rep.get_loaded_namespaces();
    for (let i = 0; i < namespaces.length; ++i) {
        let ns = namespaces[i];
        GIRepository[ns] = imports.gi[ns];
        // store the namespace name in the object so we can find it later
        // (we will make sure to delete this on extension disable())
        Object.defineProperty(GIRepository[ns], '__LG_helper_name__', {
            value: ns,
            enumerable: false,
            writable: false
        });
    }

    /* ********** Patches for shift-click with LG Inspector ********** */
    /* Desired behaviour:
     * - holding down SHIFT ungrabs the pointer such that clicking/hot corner
     *   triggering can take effect
     * - releasing SHIFT grabs the pointer back to restore the Inspector.
     *
     * We could just detect key-press Shift and call Clutter.ungrab_pointer,
     * and then listen for key-release Shift and call Clutter.grab_poiner,
     * *except* that if you select a PopupMenuItem you lose focus to the
     * looking glass causing you to get stuck not being able to input anything
     * into gnome-shell.
     * (Select item -> LG sets key focus to LG._entry -> PopupMenuManager
     * detects loss of focus to the menu -> calls .popModal() -> restores key
     * focus to whatever had focus before .pushModal (Inspector.eventHandler),
     * but this is now NULL/destroyed, so the stage gets focus instead).
     *
     * Also, since PopupMenuManager uses 'captured-event' to deal with
     * events when the menu is open, clicking on an item causes that item's
     * 'activate' method to be fired as well as it bein selected in the
     * Inspector, which is undesirable. We can't block it because 'button-press'
     * is fired *after* 'captured-event'.
     *
     * So just because of that, we instead re-write the Inspector to use
     * .pushModal and .popModal instead of grab_pointer and grab_keyboard so
     * that the modal stack isn't messed with. We also listen to
     * 'captured-event' to handle it before the PopupMenu can.
     */

    storage.Inspector = LookingGlass.Inspector.prototype;
    LookingGlass.Inspector.prototype = Inspector.prototype;

    /* ********** Patches for horizontal scrolling ********** */
    // 1. Extension descriptions should wrap.
    storage._createExtensionDisplay = EProto._createExtensionDisplay;
    EProto._createExtensionDisplay = function (meta) {
        let box = storage._createExtensionDisplay.call(this, meta),
            children = box.get_children();
        for (let i = 0; i < children.length; ++i) {
            let child = children[i];
            if (child instanceof St.Label &&
                    child.style_class === 'lg-extension-description') {
                child.clutter_text.ellipsize = Pango.EllipsizeMode.NONE;
                child.clutter_text.line_wrap = true;
            }
        }
        return box;
    };

    // 2. In the Object Inspector

    /* ********** Patches to ObjLink (copy/paste right-click menu) ********** */
    //LookingGlass.ObjLink = ObjLinkWithRightClick;
    // FIXME: the assignment to Inspector.prototype instead of Inspector is
    // just a quick hack to get things working in GNOME 3.4 (where Inspector
    // is a constant); will port properly to 3.4 later.
    storage.ObjLink = LookingGlass.ObjLink.prototype;
    LookingGlass.ObjLink.prototype = ObjLinkWithRightClick.prototype;


    /* ********** Patches to LookingGlass global functions ********** */
    // If the object has property __LG_helper_label__ we will use that as the
    // .toString() function (the reason we do not override toString() directly
    // is that then it shows up as an enumerable property of the object which we
    // don't want unless the user has caused it. Also, we don't want to
    // accidentally mess with the descriptions other users expect to get out of
    // toString().
    storage.objectToString = LookingGlass.objectToString;
    LookingGlass.objectToString = function (o) {
        // for example I want an object of functions to print with the function
        // names, but be an object underneath.
        if (o && typeof(o) === 'object') {
            let str = get_property(o, '__LG_helper_label__');
            if (str) {
                return str[1];
            }
        }
        return storage.objectToString(o);
    };

    /* ********** Patches to LookingGlass.ObjectInspector ********** */

    /* _init, open, close and _onBack are patched *just* to have a longer object
     * history (i.e. change this._previousObj to this._history).
     *
     * _init additionally stores the clipboard in it
     */
    storage._init = OIProto._init;
    OIProto._init = function () {
        storage._init.apply(this, arguments);

        // add in memory
        delete this._previousObj;
        this._history = [];

        // clipboard
        this._clipboard = St.Clipboard.get_default();
    };

    storage.open = OIProto.open;
    OIProto.open = function (sourceActor) {
        if (!this._open) {
            this._history = [];
        }
        storage.open.call(this, sourceActor);
    };

    storage.close = OIProto.close;
    OIProto.close = function () {
        if (this._open) {
            this._history = [];
        }
        storage.close.call(this);
    };

    storage._onBack = OIProto._onBack;
    OIProto._onBack = function () {
        this.selectObject(this._history.pop(), true);
    };

    /* selectObject is patched to:
     * - show function code
     * - if `obj` is a GI Namespace, makes a proxy for it with members
     *   'Functions', 'Constants', 'Classes' for the user to browse (e.g. Meta)
     * - if `obj` is a constructor with a class with .prototype, makes a link to
     *   let the user inspect the prototype object
     * - if `obj` is a class from a GI Namespace, shows the methods etc of that
     *   class (e.g. Meta.Display)
     */
    storage.selectObject = OIProto.selectObject;
    OIProto.selectObject = function (obj, skipPrevious) {
        if (!skipPrevious) {
            if (this._obj !== null) {
                //this._previousObj = this._obj;
                this._history.push(this._obj);
            }
            // truncate history to the limit (from the front)
            if (this._history.length > MAX_BACK_LEVELS) {
                this._history.splice(0, this._history.length - MAX_BACK_LEVELS);
            }
        } else if (obj === this._history[this._history.length - 1]) {
            // if skipPrevious is TRUE:
            // - if obj is the last element in _history, remove it
            // - otherwise, it's already been removed - keep _history as-is.
            //this._previousObj = null;
            this._history.pop();
        }
        /* *** This is Main.lookingGlass.selectObject *** */
        this._obj = obj;
        this._objProps = [];
        this._moreButton = new St.Button({
            label: _("More..."),
            style_class: 'lg-obj-inspector-button'
        });
        this._moreButton.connect('clicked',
            Lang.bind(this, this._showPropertyChunk));

        this._numShown = 0;

        this._container.get_children().forEach(function (child) {
            child.destroy();
        });

        let desc = obj.toString(),
            objAsStr = objectToString(obj),
            objType = typeof(obj);

        let hbox = new St.BoxLayout({ style_class: 'lg-obj-inspector-title' });
        this._container.add_actor(hbox);
        let label = new St.Label({
            text: _("Inspecting: %s: %s").format(objType, objAsStr)
        });
        label.single_line_mode = true;
        hbox.add(label, { expand: true, y_fill: false });

        // add a "copy" button
        let widget = new St.Button({
            label: _("Copy as string"),
            style_class: 'lg-obj-inspector-button'
        });
        // Copy object.toString()
        widget.connect('clicked', Lang.bind(this, this._onCopy, desc));
        hbox.add(widget);

        widget = new St.Button({
            label: _("Insert"),
            style_class: 'lg-obj-inspector-button'
        });
        widget.connect('clicked', Lang.bind(this, this._onInsert));
        hbox.add(widget);

        // @@ if (this._previousObj != null) {
        if (this._history.length) {
            widget = new St.Button({
                label: _("Back"),
                style_class: 'lg-obj-inspector-button'
            });
            widget.connect('clicked', Lang.bind(this, this._onBack));
            hbox.add(widget);
        }

        widget = new St.Button({ style_class: 'window-close' });
        widget.connect('clicked', Lang.bind(this, this.close));
        hbox.add(widget);

        /* ***** my added stuff to look for type ****** */
        let parts;
        if (objType === 'object') {
            if (desc === '[object GIRepositoryNamespace]') {
                let ns = obj.__LG_helper_name__ || guessNamespace(obj);
                // REPLACE the namespace with a (browseable) proxy for it.
                // (that means users can't do r(x).yyy for namespaces; oh well).
                if (ns) {
                    obj = createNamespaceProxy(ns, obj);
                }
            } else if (desc === '[object GIRepository]') {
                obj = GIRepository;
            } else if ((parts = INSTANCE_OR_CLASS_REGEX.exec(desc))) {
                obj = createObjectProxy(parts[1], parts[2], obj);
            } else if ((parts = INSTANCE_OR_CLASS_REGEX_PRIVATE.exec(desc))) {
                // [object _private_Meta_Rectangle] (Meta.Rectangle.prototype)
                obj = createObjectProxy(parts[1], parts[2], obj);
            } else {
                // Try to support looking up prototypes for GNOME-shell objects
                // that have been created with Lang.Class:
                // these all have a .constructor.__super__.prototype.__name__
                if (obj.constructor && obj.constructor.__super__ &&
                        obj.constructor.__super__.prototype) {
                            // add a 'Parent' item.
                }
            }
        }

        if (objType === 'function') {
            // it's something like Meta.Display, but we want to inspect
            // Meta.Display.prototype.
            if (isInterestingClass(obj)) {
                hbox = new St.BoxLayout();
                hbox.add(new ObjLink(obj.prototype,
                    _("Inspect prototype")).actor);
                this._container.add_actor(hbox);
            }

            // show full function text
            widget = new St.Label({
                text: desc
            });
            this._container.add_actor(widget);

        }
        this._obj = obj;
        /* ******* /end my added stuff to look for type ****** */
        if (objType === 'object') {
            // like Object.keys(obj) but includes methods up the prototype chain
            for (let propName in obj) {
                this._objProps.push(propName);
            }
            this._objProps.sort();
            // Try to support looking up prototypes for GNOME-shell objects
            // that have been created with Lang.Class:
            // these all have a .constructor.__super__.prototype.__name__
            if (obj.constructor && obj.constructor.__super__ &&
                    obj.constructor.__super__.prototype) {
                let theparent = obj.constructor.__super__.prototype;
                label = _("Parent");
                if (theparent.__name__) {
                    label += ' (' + theparent.__name__ + ')';
                }
                this._objProps.unshift([label, theparent]);
            }
            this._showPropertyChunk();
        } else {
            // if the text is too long and it's not an object, then show the
            // full text below. Note that .is_ellipsized() doesn't return
            // 'true' until a little bit later.
            label = label.clutter_text;
            label._id = label.connect('notify::width',
                Lang.bind(this, function (clutterText, unsure, text) {
                    if (!clutterText.realized) {
                        return;
                    }
                    clutterText.disconnect(clutterText._id);
                    if (!clutterText.get_layout().is_ellipsized()) {
                        return;
                    }

                    let fullLabel = new St.Label({
                        text: text
                    });
                    fullLabel.clutter_text.ellipsize = Pango.EllipsizeMode.NONE;
                    fullLabel.clutter_text.line_wrap = true;
                    // FIXME: this causes a few "The Actor 'actor' is
                    // currently inside an allocation cycle; calling
                    // clutter_actor_queue_relayout() is not recommended"
                    // warnings.
                    this._container.add_actor(fullLabel);
                }, objAsStr));
        }
    };

    OIProto._onCopy = function (button, clicked_button, text) {
        this._clipboard.set_text(text);
    };

    OIProto._addPropertyToInspector = function (obj, propName) {
        let link = null;
        let prop = null;
        let makelink = true;
        // if propName (_objProps[i]) is an array, then the first element is
        // the label and the second is the property.
        if (typeof(propName) === 'object') {
            prop = propName[1];
            propName = propName[0];
        } else {
            try {
                prop = obj[propName];
            } catch (e) {
                warn(_("Could not retrieve property '%s': %s").format(
                                propName, e.message));
                link = new St.Label({ text: '<error>' });
            }
        }
        // look for the __LG_helper_nolink__ property and make the object
        // link to `prop`.
        try {
            if (typeof(prop) === 'object') {
                makelink = get_property(prop, '__LG_helper_nolink__');
                // either it has the property (meaning don't make link), or
                // it doesn't have the property (so make the link)
                makelink = !makelink;
            }

            // If obj[prop] is an object with property __LG_helper_nolink__
            // being TRUE, we will just display obj[prop].widget instead of
            // an object link.
            //
            // This is a nasty hack to conveniently display (say)
            // headings in an object between all the links...
            // e.g. do 'global.screen.myextraproperty = 123;' and then
            // inspect global.screen: you will see a heading 'Other properties'
            // separating the GIR properties from the dynamically-added ones.
            //
            // We should *only* add the __LG_helper_nolink__ property to objects
            // that *we* have created, like proxies (i.e. don't add them to
            // items that will exist after this extension is disabled!).
            if (makelink) {
                link = new ObjLink(prop).actor;
            }
        } catch (e) {
            warn(_("Could not create object link for property '%s': %s").format(
                            propName, e.message));
            link = new St.Label({ text: '<error>' });
        }

        let hbox = new St.BoxLayout();
        if (!makelink) {
            hbox.add((prop.widget ||
                        new St.Label({
                            text: (prop.text || propName)
                        })
                    ));
        } else {
            hbox.add(new St.Label({ text: propName + ': ' }));
            hbox.add(link);
        }
        /* *** my stuff: if the function is a constructor, show
         * a link to the prototype (e.g.
         * Meta.Display -> Meta.Display.prototype)
         */
        if (isInterestingClass(prop)) {
            hbox.add(new St.Label({text: ' '}));
            hbox.add(new ObjLink(
                prop.prototype,
                '(' + _("Inspect prototype") + ')'
            ).actor);
        }
        /* *** /end my stuff *** */
        this._container.add_actor(hbox);
        return false;
    };

    OIProto._showPropertyChunk = function () {
        if (this._moreButton.get_parent()) {
            this._container.remove_actor(this._moreButton);
        }
        let n = Math.min(this._objProps.length,
                         this._numShown + ITEM_LIMIT);

        for (let i = this._numShown; i < n; ++i) {
            this._addPropertyToInspector(this._obj, this._objProps[i]);
            this._numShown++;
        }

        if (n !== this._objProps.length) {
            // add a 'More...' button
            this._container.add_actor(this._moreButton);
        }
    };
}

function disable() {
    /* ********** Patches for shift-click with LG Inspector ********** */
    // FIXME prototype
    LookingGlass.Inspector.prototype = storage.Inspector;

    /* ********** Patches for horizontal scrolling ********** */
    EProto._createExtensionDisplay = storage._createExtensionDisplay;

    /* ********** Patches to ObjLink (copy/paste right-click menu) ********** */
    // Future `ObjLink`s will be created as `ObjLink`s.
    // FIXME prototype
    Main.lookingGlass.ObjLink.prototype = storage.ObjLink;

    /* ********** Patches to LookingGlass global functions ********** */
    LookingGlass.objectToString = storage.objectToString;

    /* ********** Patches to LookingGlass.ObjectInspector ********** */
    OIProto.selectObject = storage.selectObject;
    OIProto._init = storage._init;
    OIProto.open = storage.open;
    OIProto.close = storage.close;
    OIProto._onBack = storage._onBack;
    delete OIProto._showPropertyChunk;
    delete OIProto._addPropertyToInspector;
    delete OIProto._onCopy;

    /* ********** Reset stuff ********** */
    // make a new object inspector for good measure (easier than me fixing
    // up ._history, ._prevObj, ._objProps, ._numShown etc)
    // This also removes all the ObjLinkWithRightClicks and replaces them
    // with ObjLinks
    Main.lookingGlass._objInspector = new LookingGlass.ObjInspector();

    storage = {};

    // remove all the __LG_helper_name__ properties we put into imports.gi.repo
    for (let ns in GIRepository) {
        if (!GIRepository.hasOwnProperty(ns)) {
            continue;
        }
        delete GIRepository[ns].__LG_helper_name__;
    }

    // note: all the other __LG_helper_xyz__ properties were not created on
    // global objects but on the proxies we created for the objects/namespaces,
    // i.e. our objects. When the objects are deleted they will be too.
}
